use crate::{builder::ServerBuilder, config::ServerConfig};
use http1::Builder;
use http_body_util::Full;
use hyper::{body::Incoming, server::conn::http1, service::service_fn, Request, Response};
use hyper_util::rt::{TokioIo, TokioTimer};
use log::error;
use std::{convert::Infallible, future::Future, net::SocketAddr, sync::Arc};
use tokio::{net::TcpListener, spawn};
use tokio_util::bytes::Bytes;

pub struct Server {
    pub config: Arc<ServerConfig>,
}

impl Server {
    pub fn builder() -> ServerBuilder {
        ServerBuilder {
            config: ServerConfig::default(),
        }
    }

    pub async fn run<F, Fut>(self, handler: F)
    where
        F: Fn(Request<Incoming>) -> Fut + Send + Sync + 'static,
        Fut: Future<Output = Result<Response<Full<Bytes>>, Infallible>> + Send,
    {
        let addr: SocketAddr = format!("{}:{}", self.config.ip, self.config.port)
            .parse()
            .expect("Invalid IP or port");
        let listener = TcpListener::bind(addr).await.unwrap();
        let handler = Arc::new(handler);

        loop {
            let listener_res = listener.accept().await;
            if listener_res.is_err() {
                continue;
            }

            let (tcp, client_addr) = listener_res.unwrap();
            let io = TokioIo::new(tcp);

            if !self.config.is_ip_authorized(&client_addr.ip()) {
                continue;
            }

            let config = Arc::clone(&self.config);
            let handler = Arc::clone(&handler);
            spawn(async move {
                if let Err(error) = Builder::new()
                    .timer(TokioTimer::new())
                    .serve_connection(
                        io,
                        service_fn(move |req| {
                            let config = Arc::clone(&config);
                            let handler = Arc::clone(&handler);

                            async move {
                                if config.is_req_authorized(&req) {
                                    handler(req).await
                                } else {
                                    Ok(Response::builder()
                                        .status(401)
                                        .body(Full::new(Bytes::from("Unauthorized")))
                                        .unwrap())
                                }
                            }
                        }),
                    )
                    .await
                {
                    error!(error = error.to_string().as_str();
                        "Serving connection"
                    );
                }
            });
        }
    }
}
